/*
 * TOFU Circuit Simulator
 * Copyright (C) 2007 Michael Birken
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
 * 
 * http://www.meatfighter.com/tofu
 */

package tofu;

import java.io.*;
import java.util.*;

public class Main {
    
  private final Node HIGH_NODE = new Node();
  private final Node CLOCK_NODE = new Node();
  private final Node ICLOCK_NODE = new Node();  
  
  private Map<String, Element> elementMap = new HashMap<String,Element>();
  private Map<String,Class> componentClassMap = new HashMap<String,Class>();
  private Element element;
  private String runLine;
  private Set<Node> nodes = new HashSet<Node>();
  private Set<Component> components = new HashSet<Component>();
  private Stack<Node> propagationStack = new Stack<Node>();  
  private boolean parsingDef;
  private boolean parsingRun;
  private boolean parsingSubelement;
  private int parameterCount;
  
  public Main() {
    addComponentClass("r", Relay.class); 
    HIGH_NODE.setVoltage(Voltage.H);
    CLOCK_NODE.setVoltage(Voltage.H);
    ICLOCK_NODE.setVoltage(Voltage.Z);
  }
  
  private void addComponentClass(String name, Class c) {
    try {
      c.forName(c.getName());
    } catch(ClassNotFoundException e) {
    }
    componentClassMap.put(name, c);
  }
  
  private Node createNode() {
    Node node = new Node();
    nodes.add(node);
    return node;
  }
  
  private void build(Element e, Map<String,Node> nodeMap) {
    for(List<String> se : e.getSubelements()) {
      String[] subelement = new String[se.size()];
      se.toArray(subelement);
      Class componentClass = componentClassMap.get(subelement[0]);
      if (componentClass == null) {
        // Component defined in input file
        Element el = elementMap.get(subelement[0]);
        if (el == null) {
          System.out.printf(
              "Error building: Can't find component %s.%n", subelement[0]);
          System.exit(0);
        }
        Map<String, Node> paramNodeMap = new HashMap<String, Node>();
        List<String> parameters = el.getParameters();
        for(int i = 1; i < subelement.length; i++) {
          Node node = "1".equals(subelement[i]) ? HIGH_NODE 
                    : "0".equals(subelement[i]) ? createNode() 
                    : "c".equals(subelement[i]) ? CLOCK_NODE
                    : "k".equals(subelement[i]) ? ICLOCK_NODE 
                    : nodeMap.get(subelement[i]);
          if (node == null) {
            node = createNode();
            nodeMap.put(subelement[i], node);
          }
          paramNodeMap.put(parameters.get(i - 1), node);
        }
        build(el, paramNodeMap);
      } else {
        // Standard component
        try {
          Component component = (Component)componentClass.newInstance();
          components.add(component);
          component.setPropagationStack(propagationStack);
          int start = 1;
          if (subelement[1].startsWith("\"") && subelement[1].endsWith("\"")
              && subelement[1].length() > 1) {
            component.setInitParam(
                subelement[1].substring(1, subelement[1].length() - 1));
            start = 2;
          }
          for(int i = start; i < subelement.length; i++) {
            Node node = "1".equals(subelement[i]) ? HIGH_NODE 
                      : "0".equals(subelement[i]) ? createNode() 
                      : "c".equals(subelement[i]) ? CLOCK_NODE
                      : "k".equals(subelement[i]) ? ICLOCK_NODE 
                      : nodeMap.get(subelement[i]);
            if (node == null) {
              node = createNode();
              nodeMap.put(subelement[i], node);
            }
            node.getComponents().add(component);
            component.putNode(component.getParameters()[i - 1], node);
          }
        } catch(Throwable t) {
          t.printStackTrace();
          System.exit(0);
        }        
      }
    }
  }
  
  private void run() {
    if (runLine == null) {
      System.out.println("Nothing to run.");
      return;
    }
    String[] tokens = split(runLine);
    for(int i = 0; i < tokens.length; i++) {
      if (i != 4) {
        tokens[i] = tokens[i].toLowerCase();
      }
    }
    
    int clockPeriod = 0;
    int iterations = 0;

    try {
      iterations = Integer.parseInt(tokens[1]);
    } catch(NumberFormatException e) {
      System.out.printf("Run error: Invalid iterations %s.%n", tokens[1]);
      return;
    }
    
    try {
      clockPeriod = Integer.parseInt(tokens[2]);
    } catch(NumberFormatException e) {
      System.out.printf("Run error: Invalid clock period %s.%n", tokens[2]);
      return;
    }
    
    Element e = null;
    Class componentClass = componentClassMap.get(tokens[3]);
    List<String> subelement = null;
    if (componentClass != null) {
      try {
        Component component = (Component)componentClass.newInstance();        
        e = new Element();
        String[] params = component.getParameters();
        e.addParameters(params, 0);
        subelement = new ArrayList<String>();
        subelement.add(tokens[3]);
        for(int i = 0; i < params.length; i++) {
          subelement.add(params[i]);
        }
        e.addSubelement(subelement);
      } catch(Throwable t) {
        t.printStackTrace();
        return;
      }
    } else {
      e = elementMap.get(tokens[3]);
    }
    if (e == null) {
      System.out.printf("Run error: Can't find component %s.%n", tokens[3]);
      return;
    }
    
    List<Node> queryNodes = new ArrayList<Node>();
    Map<String, Node> paramNodeMap = new HashMap<String, Node>();
    List<String> parameters = e.getParameters();
    if (parameters.size() != tokens.length - 4) {
      System.out.printf(
          "Run error: Wrong number of parameters for component %s.%n", tokens[3]);
      return;
    }
    for(int i = 4; i < tokens.length; i++) {
      if ("?".equals(tokens[i])) {
        Node node = createNode();
        queryNodes.add(node);
        paramNodeMap.put(parameters.get(i - 4), node);
      } else {
        if (i == 4 && subelement != null && tokens[i].startsWith("\"") 
            && tokens[i].endsWith("\"") && tokens[i].length() > 1) {
          subelement.set(1, tokens[i]);
        } else {   
          tokens[i] = tokens[i].toLowerCase();
          Node node = "1".equals(tokens[i]) ? HIGH_NODE 
              : "0".equals(tokens[i]) ? createNode() 
              : "c".equals(tokens[i]) ? CLOCK_NODE
              : "k".equals(tokens[i]) ? ICLOCK_NODE 
              : null;
          if (node == null) {
            System.out.printf(
              "Run error: Parameter \"%s\" invalid. Must be 0, 1, C, K or ?.%n", 
              tokens[i]);
            return;
          } else {
            paramNodeMap.put(parameters.get(i - 4), node);
          }
        }
      }
    }
    build(e, paramNodeMap);
    for(Component component : components) {
      component.init();
    }
    
    System.out.printf("Build complete: %d components, %d nodes.%n",
        components.size(), nodes.size());
    List<Voltage> queryVoltages = new ArrayList<Voltage>();
    StringBuilder sb = new StringBuilder();
    for(int i = 0; i < queryNodes.size(); i++) {
      Voltage voltage = queryNodes.get(i).getVoltage();
      queryVoltages.add(null);
    }
    
    int iteration = 0;
    int clockCount = clockPeriod;
    while(iterations < 1 || iteration < iterations) {
      iteration++;
      clockCount--;      
      if (clockCount == 0) {
        clockCount = clockPeriod;
        Voltage temp = CLOCK_NODE.getVoltage();
        CLOCK_NODE.setVoltage(ICLOCK_NODE.getVoltage());
        ICLOCK_NODE.setVoltage(temp);
      }
      
      for(Node node : nodes) {
        node.setVoltage(Voltage.Z);
      }
      propagationStack.clear();
      propagationStack.push(HIGH_NODE);
      propagationStack.push(CLOCK_NODE.getVoltage() == Voltage.H 
          ? CLOCK_NODE : ICLOCK_NODE);
      while(propagationStack.size() > 0) {
        Set<Component> components = propagationStack.pop().getComponents();
        for(Component c : components) {
          c.flow();
        }
      }
      
      for(Component c : components) {
        c.update();
      }
      
      for(int i = 0; i < queryNodes.size(); i++) {
        if (queryNodes.get(i).getVoltage() != queryVoltages.get(i)) {
          sb.setLength(0);
          for(int j = 0; j < queryNodes.size(); j++) {
            Voltage voltage = queryNodes.get(j).getVoltage();
            queryVoltages.set(j, voltage);
            switch(voltage) {
              case H:
                sb.append("1");
                break;
              case Z:
                sb.append("0");
                break;
            }
          }
          System.out.printf("%d: %s%n", iteration, sb.toString());
          break;
        }  
      }
    }
  }
  
  private String[] split(String str) {
    List<String> list = new ArrayList<String>();
    str = str.trim();
    
    boolean inQuote = false;
    StringBuilder sb = new StringBuilder();
    for(int i = 0; i < str.length(); i++) {
      if (str.charAt(i) == '"') {
        inQuote = !inQuote;
      }
      if ((!inQuote && Character.isWhitespace(str.charAt(i)))
          || i == str.length() - 1) {
        if (i == str.length() - 1 && !Character.isWhitespace(str.charAt(i))) {
          sb.append(str.charAt(i));
        }
        list.add(sb.toString());
        sb.setLength(0);
        do {
          i++;
        } while(i < str.length() && Character.isWhitespace(str.charAt(i)));
        i--;
      } else {
        sb.append(str.charAt(i));
      }
    }
    
    String[] strings = new String[list.size()];
    list.toArray(strings);
    return strings;
  }
  
  private boolean processConfigFileLine(int lineNumber, String line) {
    line = line.trim();
    if (line.length() == 0) {
      return true;
    }
    String[] tokens = line.split("\\s+");
    if (tokens.length < 2) {
      System.out.printf(
          "Config error on line %d: [class name] [component name] [[ ... component parameters ... ]]%n",
          lineNumber);
      return false;
    }
    String[] params = new String[tokens.length - 2];
    if (params.length > 0) {
      System.arraycopy(tokens, 2, params, 0, params.length);
    }
    Class c = null;
    try {
      c = Class.forName(tokens[0]);
    } catch(ClassNotFoundException e) {
      System.out.printf("Config error on line %d: Can't find class %s%n",
          lineNumber, tokens[0]);
      return false;
    }
    addComponentClass(tokens[1], c);
    Component.setParameters(c, params);
    return true;
  }
  
  private boolean processFileLine(int lineNumber, String line) {    

    int pivot = line.indexOf("#");
    if (pivot == 0) {
      return true;
    } else if (pivot > 0) {
      line = line.substring(0, pivot);
    }
    line = line.trim();
    if (line.length() == 0) {
      return true;
    }
    String lineWithCase = line;
    line = line.toLowerCase();
    
    if (parsingSubelement) {
      String[] tokens = split(lineWithCase);
      List<String> tokenList = new ArrayList<String>();
      for(int i = 0; i < tokens.length; i++) {
        if (elementMap.get(tokens[i]) != null 
            || componentClassMap.get(tokens[i]) != null
            || "def".equals(tokens[i]) || "run".equals(tokens[i])) {
          System.out.printf(
              "Error on line %d: Node %s matches name of component or keyword. Too few parameters on previous line?%n", 
              lineNumber, tokens[0]);
          return false;
        }         
        if (tokens[i].startsWith("\"") && tokens[i].endsWith("\"")) {
          tokenList.add(tokens[i]);
        } else {
          tokenList.add(tokens[i].toLowerCase());
        }        
      }
      if (element.addMoreParameters(tokenList) == parameterCount + 1) {
        parsingSubelement = false;
      }
      if (tokenList.size() > parameterCount + 1) {
        System.out.printf(
            "Error on line %d: Too many parameters for component %s.%n", 
            lineNumber, tokens[0]);
        return false;
      } 
    } else if (parsingDef) {
      if (line.startsWith("def") || line.startsWith("run")) {
        System.out.printf(
            "Error on line %d: Are you missing a colon in the %s def?%n", 
            lineNumber,
            element.getName());
        return false;
      }
      if (line.endsWith(":")) {
        parsingDef = false;
        line = line.substring(0, line.length() - 1);
      }
      element.addParameters(split(line), 0);
    } else if (line.startsWith("def")) {
      parsingRun = false;
      parsingDef = true;
      if (line.endsWith(":")) {
        parsingDef = false;
        line = line.substring(0, line.length() - 1);
      }
      String[] tokens = split(line);
      element = new Element();
      element.setName(tokens[1]);
      element.addParameters(tokens, 2);
      elementMap.put(element.getName(), element);
    } else if (line.startsWith("run")) {
      if (runLine != null) {
        System.out.printf(
          "Error on line %d: Muliple run commands.%n", lineNumber);
        return false;
      }
      runLine = lineWithCase;
      parsingRun = true;
      element = null;
    } else {
      if (element == null) {
        if (parsingRun) {
          runLine = runLine + " " + lineWithCase;
          return true;
        } else {
          System.out.printf(
            "Error on line %d: Subcomponent outside of def.%n", lineNumber);
          return false;
        }
      }
      parsingRun = false;
      parsingSubelement = true;
      String[] tokens = split(lineWithCase);
      List<String> tokenList = new ArrayList<String>();
      for(int i = 0; i < tokens.length; i++) {
        tokens[i] = (i == 1 && tokens[1].startsWith("\"") 
            && tokens[1].endsWith("\"") && tokens[1].length() > 1) 
                ? tokens[1] : tokens[i].toLowerCase();
        tokenList.add(tokens[i]);
        if (i > 0 && (elementMap.get(tokens[i]) != null 
            || componentClassMap.get(tokens[i]) != null
            || "def".equals(tokens[i]) || "run".equals(tokens[i]))) {
          System.out.printf(
              "Error on line %d: Node %s matches name of component or keyword.  Too few parameters on previous line?%n", 
              lineNumber, tokens[0]);
          return false;
        }
      }
      element.addSubelement(tokenList);
      Element subelement = elementMap.get(tokens[0]);
      if (subelement == null) {
        Class c = componentClassMap.get(tokens[0]);
        if (c == null) {
          System.out.printf("Error on line %d: Unknown component %s.%n", 
              lineNumber, tokens[0]);
          return false;
        } else {
          String[] parameters = Component.getParameters(c);
          parameterCount = parameters.length;          
        }
      } else {
        parameterCount = subelement.getParameters().size();
      }
      if (tokenList.size() == parameterCount + 1) {
        parsingSubelement = false;
      }
      if (tokenList.size() > parameterCount + 1) {
        System.out.printf(
            "Error on line %d: Too many parameters for component %s.%n", 
            lineNumber, tokens[0]);
        return false;
      }
    }
    return true;
  }
  
  private boolean readConfigFile(String fileName) throws IOException {
    File file = new File(fileName);
    if (!file.exists()) {
      System.out.printf("Can't find file: %s%n", fileName);
      return false;
    }
    BufferedReader br = null;
    int lineNumber = 0;
    try {
      br = new BufferedReader(new FileReader(fileName));
      String input = null;
      while((input = br.readLine()) != null) {
        if (!processConfigFileLine(++lineNumber, input)) {
          return false;
        }
      }
    } finally {
      if (br != null) {
        try {
          br.close();
        } catch(IOException e) {          
        }
      }
    }
    return true;    
  }
  
  private boolean readFile(String fileName) throws IOException {
    File file = new File(fileName);
    if (!file.exists()) {
      System.out.printf("Can't find file: %s%n", fileName);
      return false;
    }
    BufferedReader br = null;
    int lineNumber = 0;
    try {
      br = new BufferedReader(new FileReader(fileName));
      String input = null;
      while((input = br.readLine()) != null) {
        if (!processFileLine(++lineNumber, input)) {
          return false;
        }
      }
    } finally {
      if (br != null) {
        try {
          br.close();
        } catch(IOException e) {          
        }
      }
    }
    return true;
  }
  
  public static void main(String[] args) throws Throwable {
    
    if (args.length == 0 || args.length > 2) {
      System.out.println("Args: [tofu file] [optional config file]");
      return;
    }
    
    Main m = new Main();
    if (args.length == 2) {
      if (!m.readConfigFile(args[1])) {
        return;
      }
    }
    
    if (!m.readFile(args[0])) {
      return;
    }
    m.run();
  }
  
}
